@@ -32,8 +32,9 @@ class Conllu:
     def __new__(cls):
         raise NotImplementedError('This class has only static methods')
 
-    @staticmethod
-    def fix(corpus, split_multi=False, adjust_for_speech=False):
+    @classmethod
+    def fix(cls, corpus, split_multi=False, adjust_for_speech=False,
+            columns=None):
         """Fix Parsed CoNLL-U structure of *corpus* (renew token ids, generate
         sentence ids and text if they are not present in metadata, etc.)
 
@@ -41,10 +42,15 @@ class Conllu:
                             as multiword tokens
         :param adjust_for_speech: if True, remove all non alphanumeric tokens
                                   and convert all words to lower case
+        :param columns: list of column names. If None, standard CoNLL-U columns
+                        are used
+        :type columns: list(str)
         :return: sentences in Parsed CoNLL-U format
         :rtype: sequence of tuple(list(dict(str: str|OrderedDict(str: str))),
                                   OrderedDict(str: str))
         """
+        if columns is None:
+            columns = cls.STD_COLUMNS
         for sent_no, sentence in enumerate(corpus):
             sentence, sentence_meta = \
                 sentence if isinstance(sentence, tuple) else \
@@ -52,7 +58,7 @@ class Conllu:
 
             tokens = []
             id_, sub_id = 0, 0
-            multi_token, last_old_id = None, None
+            multi_token, multi_token_id, last_old_id = None, None, None
             for token in sentence:
                 if adjust_for_speech:
                     wform = token['FORM']
@@ -70,6 +76,9 @@ class Conllu:
                         sub_id += 1
                         token['ID'] = str(id_) + '.' + str(sub_id)
                     elif multi_token:
+                        assert '-' not in old_id, 'ERROR: ' \
+                            'overlapping multi-tokens {} and {}' \
+                                .format(multi_token_id, old_id)
                         id_ += 1
                         sub_id = 0
                         token['ID'] = str(id_)
@@ -84,7 +93,7 @@ class Conllu:
                             token['ID'] = str(id_)
                             dash_pos = old_id.rfind('-')
                             if dash_pos > 0:
-                                multi_token = token
+                                multi_token, multi_token_id = token, old_id
                                 last_old_id = old_id[dash_pos + 1:]
                                 id_ -= 1
                             else:
@@ -120,6 +129,8 @@ class Conllu:
                         for token_ in sentence:
                             if token_['HEAD'] == old_id:
                                 token_['HEAD'] = token['ID']
+            assert not multi_token, 'ERROR: ' \
+                "can't close multi-token {}".format(multi_token_id)
             sentence = tokens
             if not sentence:
                 vals = {}
@@ -196,12 +207,10 @@ class Conllu:
                        adjust_for_speech=False, columns=None):
         """Convert a sequence of tokenized *sentences* to Parsed CoNLL-U format
 
-        :param fix: fix CoNLL-U structure of after conversion
         :param split_multi: if True then wforms with spaces will be processed
-                            as multiword tokens (used only with fix=True)
+                            as multiword tokens
         :param adjust_for_speech: if yes, remove all non alphanumeric tokens
-                                  and convert all words to lower case (used
-                                  only with fix=True)
+                                  and convert all words to lower case
         :param columns: list of column names. If None, standard CoNLL-U columns
                         are used
         :type columns: list(str)
@@ -210,7 +219,8 @@ class Conllu:
         yield from cls.fix((cls.from_sentence(s, columns=columns)
                                for s in sentences),
                            split_multi=split_multi,
-                           adjust_for_speech=adjust_for_speech)
+                           adjust_for_speech=adjust_for_speech,
+                           columns=columns)
 
     @classmethod
     def load(cls, corpus, encoding='utf-8-sig', fix=True, split_multi=False,
@@ -294,8 +304,10 @@ class Conllu:
                         if column in ['FEATS', 'MISC']:
                             try:
                                 val = OrderedDict(
-                                    () if val == '_'
-                                       or val.startswith ('_|') else
+                                    #() if val == '_'
+                                    # fix for the last MISC feat of syntagrus
+                                    # that is '_~':
+                                    () if val.startswith ('_') else
                                     [(k, v) for k, v in [
                                         t.split('=', 1) for t in val.split('|')
                                     ]])
@@ -431,11 +443,15 @@ class Conllu:
                 print(line, end='', file=f)
 
     @classmethod
-    def merge(cls, corpus1, corpus2, encoding='utf-8-sig', stop_on_error=True,
-              log_file=LOG_FILE):
+    def merge(cls, corpus1, corpus2, encoding='utf-8-sig',
+              ignore_new_meta=False,
+              stop_on_error=True, log_file=LOG_FILE):
         """Merge CoNLL-U fields of two corpuses with identical text data.
 
-        :param stop_on_error: if `True`, the method will raise an error if
+        :param ignore_new_meta: if ``True``, output metadata will be exactly
+                                as in *corpus1*. Metadata of *corpus2* is
+                                ignored.
+        :param stop_on_error: if ``True``, the method will raise an error if
                               some field of any token has non-`None` values in
                               both corpuses and these values are not
                               equivalent. Note, that non-equivalent values in
@@ -444,61 +460,119 @@ class Conllu:
         :return: sentences in Parsed CoNLL-U format
         :rtype: sequence of tuple(list(dict(str: str|OrderedDict(str: str))),
                                   OrderedDict(str: str))
+
+        Important: if a sentence of *corpus1* has a token with an ID field
+                   that contains char '-' or '.', and a corresponding sentence
+                   of a *corpus2* doesn't contain an ID field with identical
+                   character in that position, we'd get the first token as is
+                   and compare next token of its sentence instead of it.
         """
+        sentence_meta1 = sentence_meta1 = None
+
+        def error(msg, val1, val2):
+            if not msg.endswith(':'):
+                msg += ' are not equals:'
+            msg += '\n' \
+                 + 'Value 1: "{}"\n'.format(val1) \
+                 + 'Value 2: "{}"\n'.format(val2) \
+                 + 'Meta 1: {}\n'.format(sentence_meta1) \
+                 + 'Meta 2: {}\n'.format(sentence_meta2)
+            raise ValueError(msg)
+
         corpus1 = cls.load(corpus1, encoding=encoding, fix=False,
                            log_file=log_file)
         corpus2 = cls.load(corpus2, encoding=encoding, fix=False,
                            log_file=None)
-        for sentence1, sentence2 in zip(corpus1, corpus2):
-            sentence1, sentence_meta1 = \
-                sentence1 if isinstance(sentence1, tuple) else \
-                (sentence1, OrderedDict())
-            sentence2, sentence_meta2 = \
-                sentence2 if isinstance(sentence2, tuple) else \
-                (sentence2, OrderedDict())
-
-            def error(msg, val1, val2):
-                msg += ' are not equals:\n' \
-                     + 'Value 1: "{}"\n'.format(val1) \
-                     + 'Value 2: "{}"\n'.format(val2) \
-                     + 'Meta 1: {}\n'.format(sentence_meta1) \
-                     + 'Meta 2: {}\n'.format(sentence_meta2)
-                raise RuntimeError(msg)
-
-            for meta_key, meta_val2 in sentence_meta2.items():
-                meta_val1 = sentence_meta1.get(meta_key)
-                if meta_val2 is not None:
-                    if stop_on_error and meta_val1 is not None \
-                                     and meta_val1 != meta_val2:
-                        error('Values of meta "{}"'.format(meta_key),
-                              meta_val1, meta_val2)
-                    sentence_meta1[meta_key] = meta_val2
-            for token1, token2 in zip(sentence1, sentence2):
-                for field_key, field_val2 in token2.items():
-                    field_val1 = token1.get(field_key)
-                    if field_val1 is not None and field_val2 is not None \
-                   and type(field_val1) != type(field_val2):
-                        if stop_on_error:
-                            error('Types of field "{}"'.format(field_key),
-                                  field_val1, field_val2)
-                    elif isinstance(field_val1, dict):
-                        for feat_key, feat_val2 in field_val2.items():
-                            feat_val1 = field_val1.get(feat_key)
-                            if feat_val2 is not None:
-                                if stop_on_error and feat_val1 is not None \
-                                                 and feat_val1 != feat_val2:
-                                    error('Values of feat "{}:{}"'
-                                              .format(field_key, feat_key),
-                                          feat_val1, feat_val2)
-                                field_val1[feat_key] = feat_val2
-                    else:
-                        if field_val2 is not None:
-                            if field_val1 != field_val2 and (
-                                (stop_on_error and field_val1 is not None) \
-                             or field_key == 'FORM'
-                            ):
-                                error('Values of field "{}"'
-                                          .format(field_key, feat_key),
+        for sentence2 in corpus2:
+            while True:
+                is_empty = True
+                try:
+                    sentence1 = next(corpus1)
+                except StopIteration:
+                    error('Extra sentence in corpus2:',
+                          None, sentence2)
+                sentence1, sentence_meta1 = \
+                    sentence1 if isinstance(sentence1, tuple) else \
+                    (sentence1, OrderedDict())
+                sentence2, sentence_meta2 = \
+                    sentence2 if isinstance(sentence2, tuple) else \
+                    (sentence2, OrderedDict())
+                if not ignore_new_meta:
+                    for meta_key, meta_val2 in sentence_meta2.items():
+                        meta_val1 = sentence_meta1.get(meta_key)
+                        if meta_val2 is not None:
+                            if stop_on_error and meta_val1 is not None \
+                                             and meta_val1 != meta_val2:
+                                error('Values of meta "{}"'.format(meta_key),
+                                      meta_val1, meta_val2)
+                            sentence_meta1[meta_key] = meta_val2
+                sentence1_ = sentence1
+                sentence1_ = iter(sentence1_)
+                token1 = None
+                for token2 in sentence2:
+                    id2 = token2.get('ID', '')
+                    while True:
+                        try:
+                            token1 = next(sentence1_)
+                        except StopIteration:
+                            if is_empty:
+                                break
+                            else:
+                                error('Extra token in sentence of corpus2:',
+                                      None, token2)
+                        id1 = token1.get('ID', '')
+                        if ('.' not in id1 or '.' in id2) \
+                       and ('-' not in id1 or '-' in id2):
+                            is_empty = False
+                            break
+                    if is_empty:
+                        break
+                    for field_key, field_val2 in token2.items():
+                        field_val1 = token1.get(field_key)
+                        if field_val1 is not None and field_val2 is not None \
+                       and type(field_val1) != type(field_val2):
+                            if stop_on_error:
+                                error('Types of field "{}"'.format(field_key),
                                       field_val1, field_val2)
-                            token1[field_key] = field_val2
+                        elif isinstance(field_val1, dict):
+                            for feat_key, feat_val2 in field_val2.items():
+                                feat_val1 = field_val1.get(feat_key)
+                                if feat_val2 is not None:
+                                    if stop_on_error \
+                                   and feat_val1 is not None \
+                                   and feat_val1 != feat_val2:
+                                        error('Values of feat "{}:{}"'
+                                                  .format(field_key,
+                                                          feat_key),
+                                              feat_val1, feat_val2)
+                                    field_val1[feat_key] = feat_val2
+                        else:
+                            if field_val2 is not None:
+                                if field_val1 != field_val2 and (
+                                    (stop_on_error and field_val1 is not None
+                                 and not (field_key == 'ID'
+                                      and ignore_new_meta))
+                                 or field_key == 'FORM'
+                                ):
+                                    error('Values of field "{}"'
+                                              .format(field_key),
+                                          field_val1, field_val2)
+                                if field_key != 'ID':
+                                    token1[field_key] = field_val2
+                if is_empty:
+                    yield sentence1, sentence_meta1
+                else:
+                    break
+            try:
+                next(sentence1_)
+            except StopIteration:
+                pass
+            else:
+                error('Extra token in sentence of corpus1:', token1, None)
             yield sentence1, sentence_meta1
+        try:
+            next(sentence1_)
+        except StopIteration:
+            pass
+        else:
+            error('Extra sentence in corpus1:', sentence1, None)
